# models to generate different environments that the car can drive in.
# This includes terrain, clouds, wind, solar conditions, and the route along the terrain.


import jax
import jax.numpy as jnp
from jax import random
import pyqtgraph as pg
from functools import partial
from pyqtgraph.Qt import QtCore, QtGui
from typing import NamedTuple
import matplotlib.pyplot as plt
import sys


class TerrainParams(NamedTuple):
    size: int = 256
    octaves: int = 6
    persistence: float = 0.5
    lacunarity: float = 2.0
    seed: int = 42


def lerp(a, b, t):
    # assume a and b are pairs of numbers
    x = jnp.array([0,1])
    f = jnp.array([a,b])
    return jnp.interp(t, x, f)

# @partial(jax.jit, static_argnums=(2,))
# def _make_noise_layer(key: random.PRNGKey, frequency: float, shape) -> jnp.ndarray:
#
#     noise = random.normal(key, shape)
#     # create the grid.
#     x = jnp.linspace(0, shape[0] - 1, )

import jax
import jax.numpy as jnp

def generate_permutation():
    """Generate a permutation table."""
    p = jnp.arange(256, dtype=jnp.int32)
    return jnp.concatenate([p, p])

@jax.jit
def fade(t):
    """Fade function for smooth interpolation."""
    return t * t * t * (t * (t * 6 - 15) + 10)

@jax.jit
def lerp(t, a, b):
    """Linear interpolation."""
    return a + t * (b - a)

@jax.jit
def grad(hash, x, y):
    """Calculate gradient."""
    h = hash & 15
    grad_x = jnp.where(h < 8, x, y)
    grad_y = jnp.where(h < 4, y, jnp.where((h == 12) | (h == 14), x, y))
    return jnp.where(h & 1, -grad_x, grad_x) + jnp.where(h & 2, -grad_y, grad_y)

def perlin(pos):
    """ Perlin noise. Shape (N) where N = n_dims (2,3) """
    
    cellpos = pos % 1.0 # get the position inside the cell

    upos = fade(pos)

@jax.jit
def perlin_noise_2d(x, y, p):
    """Generate 2D Perlin noise value."""
    # Floor coordinates
    xi = jnp.floor(x).astype(jnp.int32) & 255
    yi = jnp.floor(y).astype(jnp.int32) & 255
    
    # Fractional coordinates
    xf = x - jnp.floor(x)
    yf = y - jnp.floor(y)
    
    # Fade curves
    u = fade(xf)
    v = fade(yf)
    
    # Hash coordinates of cube corners
    aa = p[p[xi] + yi]
    ab = p[p[xi] + yi + 1]
    ba = p[p[xi + 1] + yi]
    bb = p[p[xi + 1] + yi + 1]
    
    # Gradients
    g1 = grad(aa, xf, yf)
    g2 = grad(ba, xf - 1, yf)
    g3 = grad(ab, xf, yf - 1)
    g4 = grad(bb, xf - 1, yf - 1)
    
    # Interpolate
    x1 = lerp(u, g1, g2)
    x2 = lerp(u, g3, g4)
    return lerp(v, x1, x2)

def generate_noise_grid(width, height, scale=50.0):
    """Generate a grid of Perlin noise values."""
    p = generate_permutation()
    # compute the gradient grid.
    gradgrid = 
    x = jnp.linspace(0, width/scale, width)
    y = jnp.linspace(0, height/scale, height)
    X, Y = jnp.meshgrid(x, y)
    return perlin_noise_2d(X, Y, p)

# Example usage:
key = jax.random.PRNGKey(23)
noise = generate_noise_grid(256, 256)
plt.imshow(noise)
plt.savefig("output.png")


def GymV1():
    """ Makes a version 1 gym - simply an elevation profile. """

